import os
import pty
import sys
import select
import subprocess
import json

def start_training(
    model_path,
    dataset_path,
    recipe_path,
    recipe_id,
    recipe_revision,
    run_path,
    client_params,
    adapter_name,
):
    run_file_path = os.path.join(os.path.dirname(__file__), "run.py")

    # Create command list
    command = [
        "deepspeed",
        run_file_path,
        "--model_path", model_path,
        "--dataset_path", dataset_path,
        "--recipe_path", recipe_path,
        "--run_path", run_path,
        "--client_params", json.dumps(client_params),
        "--adapter_name", adapter_name,
        "--recipe_id", recipe_id,
        "--recipe_revision", recipe_revision,
    ]

    # Create pseudo-terminal
    master, slave = pty.openpty()

    # Start the subprocess with the allocated TTY
    process = subprocess.Popen(
        command,
        stdin=slave,
        stdout=slave,
        stderr=slave,
        env={**os.environ, "PYTHONUNBUFFERED": "1"},
        close_fds=True
    )

    # Close slave in parent process
    os.close(slave)

    try:
        while process.poll() is None:
            # Non-blocking check for output
            ready, _, _ = select.select([master], [], [], 0.1)
            
            if master in ready:
                # Read raw bytes from TTY
                try:
                    output = os.read(master, 1)
                    # Write raw bytes to stdout (preserving \r)
                    #check if hasattr buffer
                    if hasattr(sys.stdout, "buffer"):
                        sys.stdout.buffer.write(output)
                        sys.stdout.flush()
                    else:
                        sys.stdout.write(output)
                    
                except OSError:
                    # Handle broken pipe or other OS errors
                    pass
                    
    finally:
        # Clean up
        os.close(master)
        
        # Ensure process is terminated
        if process.poll() is None:
            process.terminate()
            try:
                process.wait(timeout=5)
            except subprocess.TimeoutExpired:
                process.kill()
                process.wait()

    return process.returncode